/* ipaudit.c
 *
 * ipaudit - network traffic data gathering
 * By Jon Rifkin <j.rifkin@uconn.edu>
 * Copyright 1999 Jonathan Rifkin
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


/*
------------------------------------------------------------------------
Compile Switches
------------------------------------------------------------------------
*/
#define DEBUG

#define DUMP
#undef  DUMP

#define HASH
#undef  HASH


/*
------------------------------------------------------------------------
Include Files
------------------------------------------------------------------------
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <error.h>
#include <pcap.h>
#include <signal.h>
#include <sys/types.h>
#include <limits.h>
#include <unistd.h>
#include "hash.h"



/*
------------------------------------------------------------------------
Defines
------------------------------------------------------------------------
*/
#define VERSION_STR "ipaudit 0.91"

#define TRUE 1
#define FALSE 0

#if 0
/*  TCP ports to dump  */
#define NPORT           6

#define P_FTP          21
#define P_TELNET       23
#define P_SUNRPC      111
#define P_LOGIN       513
#define P_INGRESLOCK 1524
#define P_IRC        6667

/*  Misc protocols to dump  */
#define PROT_ICMP       1
#endif

/*  Protocols with port info  */
#define PROT_TCP        6
#define PROT_UDP       17


/*  Flags for udp/tcp accepting all/some ports  */
#define PROT_ACC_ALL    1
#define PROT_ACC_SOME   2

/*  Length of saved packets  */
#define PLEN_DEF 96  /* default  */
#define PLEN_MIN 68  /* min allowed  */

/*  
Length of hash slots
NOT number of packets, they're unlimited (except for memory)
*/
#define NHSLOTS 1000000

/*  Number of 1/10,000 of second in second  */
#define M0SEC 10000

/*
------------------------------------------------------------------------
DEBUGGING MACROS
------------------------------------------------------------------------
*/
#ifdef DEBUG
#define WRITEMSG \
   printf ("File %s line %d: ", __FILE__, __LINE__); \
   printf ("errmsg <%s>\n", strerror(errno)); fflush(stdout);
#define WRITEVAR(VAL,FMT) \
   printf ("File %s line %d: ", __FILE__, __LINE__); \
   printf ("%s=",#VAL); printf (#FMT, VAL); printf ("\n"); \
   fflush(stdout);
#else
#define WRITEMSG
#define WRITEVAR(VAL,FMT)
#endif



/*
------------------------------------------------------------------------
MACROS
------------------------------------------------------------------------
*/
/*  Convert time in 1/M0SEC to hours, min, seconds, 1/M0SEC  */
#define HMS(hour,min,sec,sec4) \
	sec   = sec4/M0SEC; \
	sec4 -= M0SEC*sec; \
	min   = sec/60; \
   sec  -= 60*min; \
   hour  = min/60; \
   min  -= 60*hour;


/*
------------------------------------------------------------------------
Type Definitions
------------------------------------------------------------------------
*/

/*  Packet structure used by pcap library  */
typedef struct {
   u_char src[6];
   u_char dst[6];
   u_char ptype[2];     /*  ==0x800 if ip  */
   u_char version[1];
   u_char service[1];
   u_char length[2];
   u_char id[2];
   u_char flag[2];
   u_char ttl[1];
   u_char prot[1];
   u_char chksum[2];
   u_char srcip[4];
   u_char dstip[4];
   u_char srcpt[2];
   u_char dstpt[2];
   } pkt_struct_t;



/*  Start and stop time of each connection  */
typedef struct {
	/*  Time (in sec/10,000) of first and last packet  */
	int first_time, last_time;
	/*  Indentity of machine source for first, last packet  */
	unsigned char first_mach, last_mach;
} datatime_t;


/*  All data for connection  */
typedef struct {
	int        nbyte1, nbyte2, npkt1, npkt2;
	datatime_t time;
} data_t;







/*
------------------------------------------------------------------------
Global variables
------------------------------------------------------------------------
*/
extern int errno;



/*
------------------------------------------------------------------------
Module variables
------------------------------------------------------------------------
*/
int isig_m=0;

u_char *prots_m = NULL;

u_char *tcp_ports_m = NULL;
u_char *udp_ports_m = NULL;


/*  Flag for writing connection time in output  */
int write_time_m = FALSE;

int  debug_m     = FALSE;

/*
------------------------------------------------------------------------
Local Function Prototypes
------------------------------------------------------------------------
*/
void ihandler (int);
void dum2 (pkt_struct_t *);
void dum1 (struct pcap_pkthdr *);
void dum3 (hlist_t **);
void storepkt (struct pcap_pkthdr *, pkt_struct_t *, hlist_t **);
void writepkttxt (hlist_t **, char *, char *);
void writepktbin (hlist_t **, char *, char *);
void PrintUsage();

void parse_portstr(char *str);
void add_protocol(int val);
void add_port (int prot, int port);

void *alloc_err (int, int);
char *get_localstr (char *);

int  get_pkttime (struct pcap_pkthdr *);



/*
------------------------------------------------------------------------
Main Function
------------------------------------------------------------------------
*/
int main (int argc, char *argv[]) {
   pcap_t *infile;
   char   ebuf[PCAP_ERRBUF_SIZE];
   struct pcap_pkthdr pkthdr;
   pkt_struct_t *pkt;
   struct bpf_program  fcode;
   pcap_dumper_t *df = NULL;
   int dump_this = FALSE;
   int nethpkt;
   int nippkt;
   int ival;
   int length;
   int i;
   u_char *dumptable = NULL;
   char *cmd = "ip";
   int  optchar;
   int  dumpall    = FALSE;
	int  outputbin  = FALSE;
   FILE *pidfile   = NULL;
   char *progarg   = NULL;
   char *progfile  = NULL;
   char *writefile = NULL;
   char *readfile  = NULL;
   char *outfile   = NULL;
   int  maxpkt     = 0;
	int  nlen       = PLEN_DEF;   /*  Packet length to dump  */
	char *localstr  = NULL;			/*  Sorting order for ip address */
   /*  Hash table for ip connections  */
   hlist_t **hconn = NULL;
   



   /*  Read options  */
   while (-1 != (optchar=getopt(argc,argv,"abc:di:f:l:o:p:r:s:tx:w:v"))) {
      switch (optchar) {
      case '?':
         return 1;
		case 'v':
			printf ("%s (compiled %s)\n", VERSION_STR, __DATE__);
			return 0;
      case 'a':
         dumpall = TRUE;
         break;
		/*  Output packet statistics in binary file format  */
		case 'b':
			outputbin = TRUE;
			break;
		/*  Debugging option  */
		case 'd':
			debug_m = TRUE;
			break;
      /*  Write pid file  */
      case 'i':
         pidfile = fopen(optarg,"wt");
         if (NULL==pidfile) {
            fprintf (stderr, "ERROR:  Cannot open pidfile '%s'\n", optarg);
            exit(1);
         }
         fprintf (pidfile, "%d\n", getpid());
         fclose (pidfile);
         break;
      case 'x':
         progfile = strdup(optarg);
         break;
      case 'r':
         readfile = strdup(optarg);
         break;
      case 'w':
         writefile = strdup(optarg);
         break;
      case 'c':
         maxpkt    = atoi(optarg);
         break;
      case 'o':
         outfile   = strdup(optarg);
         break;
      case 'p':
         parse_portstr(strdup(optarg));
         break;
		case 's':
			nlen = atoi(optarg);
			if (nlen<PLEN_MIN)  
				nlen = PLEN_MIN;
			break;
		case 'l':
			localstr = get_localstr (optarg);
			break;
		case 't':
			write_time_m = TRUE;
			break;
      default:
         return 1;
      }
   }
   
   
   
   /*  If not reading pcap file need the interface name  */
   if (!readfile && argc-optind<1)  {
      PrintUsage();
      return 0;
   }


   

   /*  Open pcap file  */
   if (readfile)
      infile = pcap_open_offline(readfile, ebuf);
   else {
      infile = pcap_open_live(argv[optind], nlen, 1, 1000, ebuf);
      if (infile==NULL) {
         printf("ipaudit: %s (%s)\n", ebuf, "Do you need root?");
         return 1;
      }
      if (pcap_compile(infile, &fcode, cmd, 0, 0) < 0)
         printf("compile: %s", pcap_geterr(infile));
      if (pcap_setfilter(infile, &fcode) < 0)
         printf("setfilter:  %s", pcap_geterr(infile));
   }



   /*  Install interupt handler if reading live */
   if (!readfile) {
      signal (SIGINT,    ihandler);   /*  intercepts  ^C           */
      signal (SIGTERM,   ihandler);   /*  intercepts  ^kill <PID>  */
   }

   /*  Initialize hash table  */
   hconn = init_htable(NHSLOTS);


   /*  Counters  */
   nethpkt = 0;
   nippkt = 0;

   /*  Open dump file  */
   if (writefile) df = pcap_dump_open(infile, writefile);

   /*  Read inteface until interupt signal  */
   while (isig_m==0) {
      nethpkt++;
      pkt = (pkt_struct_t *) pcap_next (infile, &pkthdr);

      /*  If pkt is null and we're reading file then we're done  */
      /*  Otherwise if reading live try again                    */
      if (pkt==NULL) {
         if (readfile) break;
         else          continue;
      }

      /*  Don't exceed limit of ip packets  */
      nippkt++;
      if (maxpkt && nippkt>maxpkt)
         break;

      /*  Set ports to 0 if not UDP or TCP  */
      if ( pkt->prot[0]!=0x11 && pkt->prot[0]!=0x06 ) {
         pkt->srcpt[0] = 0;
         pkt->srcpt[1] = 0;
         pkt->dstpt[0] = 0;
         pkt->dstpt[1] = 0;
      }

      /*  Store packets  */
      storepkt (&pkthdr, pkt, hconn);


      /*  Using dump  */
      if (writefile) {

         /*  User specified to dump all packets  */
         if (dumpall) {
            dump_this = TRUE;

         } else {
            /*  See if this protocol is dumped  */
            dump_this = prots_m[pkt->prot[0]];

            /*  If udp or tcp, are ports specified ?  */
            if (PROT_ACC_SOME==dump_this) {
               dumptable = (PROT_TCP==pkt->prot[0]) ? tcp_ports_m : udp_ports_m;
               dump_this = 
                  dumptable[(pkt->srcpt[0]<<8)+pkt->srcpt[1]] ||
                  dumptable[(pkt->dstpt[0]<<8)+pkt->dstpt[1]];
            }
         }

         /*  Dump packet  */
         if (dump_this)
            pcap_dump ((u_char *) df, &pkthdr, (u_char *) pkt);
      }

      


   }  /*  Read tcpdump data  */


   /*  Close dump file  */
   if (writefile) pcap_dump_close (df);

   /*  Close packet input */
   pcap_close (infile);


   /*  Write packet info  */
	if (outputbin) {
	   writepktbin(hconn, outfile, localstr);
	} else {
	   writepkttxt(hconn, outfile, localstr);
	}

   /*  Call next program  */
   if (progfile) {
      progarg = index (progfile, ' ');
      if (progarg) {
         *progarg++ = '\0';
      }
      execl (progfile, progfile, progarg, NULL);
   }
}



/*
Interupt handler (called when program recieves operating system signal
*/
void ihandler (int cursig)
   {

   /*  SET SIGNAL FLAG  */
   isig_m = 1;

   /*  FLUSH BUFFERS  */
   fflush (stdout);

   /*  RE-INSTALL SIGNAL HANDLER  */
   signal (cursig, ihandler);
#ifdef DEBUG
	if (debug_m)
		fprintf (stderr, "ipaudit received signal number <%i>\n", cursig);
#endif
   }


/*
Store packet info in hash table, 
keyed by ip1,ip2,port1,por2,protocol
data  is number of incoming/outgoing bytes, packets
*/
void storepkt (struct pcap_pkthdr *pkthdr, pkt_struct_t *p, hlist_t **ha) {
   u_char     key[13];
   data_t     *data;
   int        ndata;
   int        length;
	data_t     idata;
	datatime_t idatatime;
	int        datasize;

   /*  Calculate data packet length  */
   length = p->length[1] + 256*(int) p->length[0] + 14;

   /*  
	Make key - order so smallest ip first 
	store data
	*/
   if (memcmp(p->srcip, p->dstip, 4) < 0) {
      memcpy (key+ 0, p->srcip, 4);
      memcpy (key+ 4, p->dstip, 4);
      memcpy (key+ 8, p->srcpt, 2);
      memcpy (key+10, p->dstpt, 2);
		idata.nbyte1 = 0;
		idata.nbyte2 = length;
		idata.npkt1  = 0;
		idata.npkt2  = 1;
   } else {
      memcpy (key+ 0, p->dstip, 4);
      memcpy (key+ 4, p->srcip, 4);
      memcpy (key+ 8, p->dstpt, 2);
      memcpy (key+10, p->srcpt, 2);
		idata.nbyte1 = length;
		idata.nbyte2 = 0;
		idata.npkt1  = 1;
		idata.npkt2  = 0;
   }
   memcpy (key+12, p->prot,  1);

	/*  Store time if requested  */
	if (write_time_m) {
		idata.time.first_time = get_pkttime (pkthdr);
		/*  If machine 1 received packet, then source was 2, else 1  */
      idata.time.first_mach = (idata.npkt1==1) ? 2 : 1;
		/*  Set first and last machine info the same  */
		idata.time.last_time  = idata.time.first_time;
		idata.time.last_mach  = idata.time.first_mach;
		/*  Set size of full data structure  */
		datasize = sizeof(data_t);

	/*  Set datasize to not store time (saves memory space)  */
	} else {
		datasize = sizeof(data_t) - sizeof(datatime_t);
	}

#ifdef DUMP
printf ("%03u.%03u.%03u.%03u ", key[0], key[1], key[2], key[3]);
printf ("%03u.%03u.%03u.%03u ", key[4], key[5], key[6], key[7]);
printf ("%u %u %u %d %d\n", key[12], (int )key[8]*256+key[9], (int )key[10]*256+key[11], length, 1);
#endif


   /*  Add to table  */
	if (! finddata(ha,(u_char *)&key,sizeof(key),(u_char **)&data,&ndata) ) {
		datasize = sizeof(idata);
      addnode (ha, (u_char *) &key, sizeof(key), (u_char *) &idata, datasize);
   } else {
		data->nbyte1 += idata.nbyte1;
		data->nbyte2 += idata.nbyte2;
		data->npkt1  += idata.npkt1;
		data->npkt2  += idata.npkt2;
		/*  Update last packet time  */
		if (write_time_m) {
			data->time.last_time  = idata.time.last_time  ;
			data->time.last_mach  = idata.time.last_mach;
		}
#ifdef DUMP
printf ("data idata  <%u %u %u %u>a  <%u %u %u %u>\n", 
data[0], data[1], data[2], data[3],
idata[0], idata[1], idata[2], idata[3]);
#endif
   }
}


/*
Retrieve and print packets from hash table in bin format
*/
void writepktbin (hlist_t **ha, char *outname, char *localstr) {
   hlist_t *t;
   data_t  *data;
   FILE    *outfile = stdout;
	char    ip1[16], ip2[16];
	int     switch_mach;
	int     first_mach, last_mach;

   if (outname) 
      outfile = fopen (outname, "wb");

   if (NULL==outfile) {
      fprintf (stderr, "ERROR:  Cannot open output file <%s>\n", outname);
      exit(1);
   }

   /*  Walk list  */
   getfirst (ha);
   while ((t=getnext(ha))) {

		/*  Get ip addresses and ports  */
		sprintf (ip1, "%03u.%03u.%03u.%03u", 
			t->key[0], t->key[1], t->key[2], t->key[3]);
		sprintf (ip2, "%03u.%03u.%03u.%03u", 
			t->key[4], t->key[5], t->key[6], t->key[7]);
		data = (data_t *) t->data;

/*test*/
printf ("Data <%d> <%d> <%d> <%d>\n", data[0], data[1], data[2], data[3]);
		/*  Re-order ip addresses if 2nd is local and first is not  */
		switch_mach = ( 
			localstr && 
		 	 strncmp(localstr,ip1,strlen(localstr)) &&
			!strncmp(localstr,ip2,strlen(localstr)) );
		if (switch_mach) {
		
			fwrite (t->key+4, 1, 4, outfile);   /* 2nd ip  */
			fwrite (t->key  , 1, 4, outfile);   /* 1st ip  */
			fwrite (t->key+12,1, 1, outfile);   /* protocol */
			fwrite (t->key+10,1, 2, outfile);   /* 2nd port  */
			fwrite (t->key+8, 1, 2, outfile);   /* 1st port  */
			fwrite (&data->nbyte2, 4, 1, outfile);  /* 2nd ip, bytes received  */ 
			fwrite (&data->nbyte1, 4, 1, outfile);  /* 1st ip, bytes received  */
			fwrite (&data->npkt2,  4, 1, outfile);  /* 2nd ip, packets recevied  */
			fwrite (&data->npkt1,  4, 1, outfile);  /* 1st ip, packets received  */

		} else {
		
			fwrite (t->key  , 1, 4, outfile);   /* 1st ip  */
			fwrite (t->key+4, 1, 4, outfile);   /* 2nd ip  */
			fwrite (t->key+12,1, 1, outfile);   /* protocol */
			fwrite (t->key+8, 1, 2, outfile);   /* 1st port  */
			fwrite (t->key+10,1, 2, outfile);   /* 2nd port  */
			fwrite (&data->nbyte1, 4, 1, outfile); /* 1st ip, bytes received  */
			fwrite (&data->nbyte2, 4, 1, outfile); /* 2nd ip, bytes received  */
			fwrite (&data->npkt1,  4, 1, outfile); /* 1st ip, packets received  */
			fwrite (&data->npkt2,  4, 1, outfile); /* 2nd ip, packets recevied  */
		}

		/*  If switching machine order, correct first/last machine id  */
		if (write_time_m) {
			if (switch_mach) {
				first_mach  = 3 - data->time.first_mach;
				last_mach   = 3 - data->time.last_mach;
			} else { 
				first_mach  = data->time.first_mach;
				last_mach   = data->time.last_mach;
			}
			fwrite (&first_mach, sizeof(first_mach), 1, outfile);
			fwrite (&last_mach, sizeof(last_mach), 1, outfile);
		}
   }

   /*  Close file  */
   if (outname)
      fclose(outfile);
}


/*
Retrieve and print packets from hash table in text format
*/
void writepkttxt (hlist_t **ha, char *outname, char *localstr) {
   hlist_t *t;
   data_t  *data;
   FILE    *outfile = stdout;
	char    ip1[16], ip2[16];
	int     pt1, pt2, prot;
	int     hour,min,sec,msec;
	int     switch_mach;
	int     first_mach, last_mach;

   if (outname) 
      outfile = fopen (outname, "wt");

   if (NULL==outfile) {
      fprintf (stderr, "ERROR:  Cannot open output file <%s>\n", outname);
      exit(1);
   }

   /*  Walk list  */
   getfirst (ha);
   while ((t=getnext(ha))) {

		/*  Get ip addresses and ports  */
		sprintf (ip1, "%03u.%03u.%03u.%03u", 
			t->key[0], t->key[1], t->key[2], t->key[3]);
		sprintf (ip2, "%03u.%03u.%03u.%03u", 
			t->key[4], t->key[5], t->key[6], t->key[7]);
      pt1  = (int) t->key[ 8]*256 + t->key[ 9];
      pt2  = (int) t->key[10]*256 + t->key[11];
		prot = t->key[12];

		/*  Re-order ip addresses if 2nd is local and first is not  */
		switch_mach = (
			localstr && 
		 	 strncmp(localstr,ip1,strlen(localstr)) &&
			!strncmp(localstr,ip2,strlen(localstr)) );

		if (switch_mach) {
		
			/*  Print key info  */
			fprintf (outfile, "%s %s %u %u %u", ip2, ip1, prot, pt2, pt1);

			/*  Data  */
			data = (data_t *) t->data;
			fprintf (outfile, "  %d %d %d %d", 
				data->nbyte2, data->nbyte1, data->npkt2, data->npkt1);

		} else {
		
			/*  Print key info  */
			fprintf (outfile, "%s %s %u %u %u", ip1, ip2, prot, pt1, pt2);

			/*  Data  */
			data = (data_t *) t->data;
			fprintf (outfile, "  %d %d %d %d", 
				data->nbyte1, data->nbyte2, data->npkt1, data->npkt2);
		}

		if (write_time_m) {
			/*  Convert seconds from midnight to 24 hour time  */
			msec  = data->time.first_time;
			HMS(hour,min,sec,msec)
			fprintf (outfile, " %02d:%02d:%02d.%04d", hour,min,sec,msec);
			msec  = data->time.last_time;
			HMS(hour,min,sec,msec)
			fprintf (outfile, " %02d:%02d:%02d.%04d", hour,min,sec,msec);

		   /*  If switching machine order, correct first/last machine id  */
			if (switch_mach) {
				first_mach  = 3 - data->time.first_mach;
				last_mach   = 3 - data->time.last_mach;
			} else { 
				first_mach  = data->time.first_mach;
				last_mach   = data->time.last_mach;
			}


			fprintf (outfile, " %1d %1d", first_mach, last_mach);
		} 

		fprintf (outfile, "\n");

#if 0
      /*  IP addresses  */
      fprintf (outfile, "%03u.%03u.%03u.%03u ", 
			t->key[0], t->key[1], t->key[2], t->key[3]);
      fprintf (outfile, "%03u.%03u.%03u.%03u ", 
			t->key[4], t->key[5], t->key[6], t->key[7]);
      /*  Protocol  */
      fprintf (outfile, "%u ", t->key[12]);
      /*  Ports      */
      fprintf (outfile, "%u ", (int) t->key[ 8]*256 + t->key[ 9]);
      fprintf (outfile, "%u ", (int) t->key[10]*256 + t->key[11]);
#endif
   }

   /*  Close file  */
   if (outname)
      fclose(outfile);
   
};


void PrintUsage(void) {
   printf ("\nUsage: ipaudit -acioprwx [interface]\n");
   printf ("  Read and record info on ip connections and optionally\n");
   printf ("  dump packets to file\n");
   printf ("\n");
   printf ("  -r readfile   -  Read packets from pcap format file\n");
   printf ("                   Don't need interface with this option\n");
   printf ("  -w writefile  -  Dump selected packets to pcap format file\n");
   printf ("                   MUST also supply either -a or -p options\n");
   printf ("  -a            -  Dump ALL packets to writefile.\n");
   printf ("                   Incompatible with -p option\n");
   printf ("  -i pidfile    -  Write process id to file\n");
	printf ("  -l net-ip     -  Order output ip address pairs by network ip\n");
   printf ("  -s nlen       -  Dump first <nlen> bytes of each packet (default %d, min %d)\n",
		PLEN_DEF, PLEN_MIN);
   printf ("  -x program    -  Run program when done\n");
   printf ("  -c npacket    -  Only read in specific number of ip packets\n");
   printf ("  -o outfile    -  Place output in 'outfile'. Default is stdout\n");
   printf ("  -p string     -  Dump only selected ip protocols and ports\n");
   printf ("     string format -p n:n:n,p,p:n where n is protocol number\n");
   printf ("     and p is port number (only for protocols 6 (tcp) and 17 (udp)\n");
   printf ("     Incompatible with -a option.\n");
	printf ("  -v            -  Print version and exit.\n");
	printf ("  -b            -  Write output in binary format (experimental)\n");
	printf ("  -t            -  Write connection start/end time\n");
   printf ("Example:\n");
   printf ("  ipaudit -w dump.fil -p1:2:6,21,23  eth0\n");
   printf ("Write only packets with protocols 1 (icmp), 2 (?), and 6 (tcp)\n");
   printf ("and tcp ports 21 and 23 (ftp,telnet)\n\n");
}



/*
Determine protocols and ports to accept from string of format
   "1:6,21,23,6667:17" means protocols 1 (icmp), 6 (tcp) and 17 (udp)
   For tcp only ports 21,23,6667.
   For udp all ports (default)
*/
void parse_portstr(char *str) {
   char *p;
   char delim;
   int  lastprot = -1;
   int  is_prot;
   int  val;

   is_prot = TRUE;
   delim=' ';
   while (delim) {
      /*  Find : or '\0'  */
      p=str;
      /*  Find next : , '\0'  */
      while (*p!=':' && *p!=',' && *p!='\0') 
         p++;
      delim = *p;
      *p = '\0';
      val = atoi(str);
      str = p+1;

      if (is_prot) {
         lastprot = val;
         add_protocol(val);
      } else {
         add_port (lastprot,val);
      }
      is_prot = (delim==':');
   }
}


void add_protocol(int val) {
   /*  Check for valid protocol  */
   if (val<0 || val>255)
      return;
   /*  ALlocate if not done already  */
   if (NULL==prots_m)
      prots_m = (u_char *) alloc_err(256, sizeof(u_char));
   if (NULL==prots_m) {
      fprintf (stderr, "ERROR: Out of memory\n");
      exit (1);
   }
   /* Accept all ports  */
   prots_m[val] = PROT_ACC_ALL;
   }


void add_port (int prot, int port) {
   /*  Check for valid port  */
   if (prot<0 || prot>0x0ffff)
      return;
   /*  Check for valid protocol  */
   if (prot==PROT_TCP) {
      prots_m[prot] = PROT_ACC_SOME;
      if (NULL==tcp_ports_m) 
         tcp_ports_m = (u_char *) alloc_err(0x010000, sizeof(u_char));
      tcp_ports_m[port] = 1;
   } else if (prot==PROT_UDP) {
      prots_m[prot] = PROT_ACC_SOME;
      if (NULL==udp_ports_m) 
         udp_ports_m = (u_char *) alloc_err(0x010000, sizeof(u_char));
      tcp_ports_m[port] = 1;
   }
}



void *alloc_err (int n, int size) {
   void *p;
   p = calloc (n, size);
   if (NULL==p) {
      fprintf (stderr, "ERROR: Out of memory.\n");
      exit(1);
   }
   return p;
}


/*  
Read and format ip address :
  for example,  "112.51.1" -> "112.051.001"
*/
char *get_localstr(char *instr) {
   /*  Room for string containing formated ip address  */
	char *buffer = (char *) calloc(16,sizeof(char)); 
	char *b = buffer;
	char triplet[3];
	char *p = instr;
	int  i,n;

	/*  Search over all bytes (things between ..)  */
	while (1) {
		n = 0;
		while ((*p>='0' && *p<='9') && n<3) 
			triplet[n++] = *p++;
		/*  Either non-digit or two many digits - return NULL */
		if (*p!='.' && *p!=0) return NULL;
		for (i=n;i<3;i++) *b++ = '0';
		for (i=0;i<n;i++) *b++ = triplet[i];
		*b++ = *p;
		if (*p==0)	return buffer;
		p++;
	}
}


/*
Convert time from pcap pkthdr format (double longs) to 
1/10,000 seconds since midnight
*/
int get_pkttime (struct pcap_pkthdr *pkthdr) {
	struct tm *time;
	time = localtime(&pkthdr->ts.tv_sec);
	return  
		(int)
		pkthdr->ts.tv_usec/100 +  /*  Convert microseconds */
		M0SEC * (time->tm_sec + 60*(time->tm_min + 60*time->tm_hour));
}
